@@ -8,6 +8,8 @@ from PyQt5.QtCore import pyqtSignal, QObject, QDir, Qt
 from PyQt5.QtWidgets import QApplication
 
 import urh.cythonext.signal_functions as signal_functions
+
+from urh import settings
 from urh.ainterpretation import AutoInterpretation
 from urh.signalprocessing.Filter import Filter
 from urh.signalprocessing.IQArray import IQArray
@@ -43,6 +45,7 @@ class Signal(QObject):
         self.__samples_per_symbol = 100
         self.__pause_threshold = 8
         self.__message_length_divisor = 1
+        self.__costas_loop_bandwidth = 0.1
         self._qad = None
         self.__center = 0
         self._noise_threshold = 0
@@ -51,6 +54,8 @@ class Signal(QObject):
         self.noise_max_plot = 0
         self.block_protocol_update = False
 
+        self.iq_array = IQArray(None, np.int8, 1)
+
         self.wav_mode = filename.endswith(".wav")
         self.__changed = False
         if modulation is None:
@@ -61,6 +66,8 @@ class Signal(QObject):
 
         self.__parameter_cache = {mod: {"center": None, "samples_per_symbol": None} for mod in self.MODULATION_TYPES}
 
+        self.__already_demodulated = False
+
         if len(filename) > 0:
             if self.wav_mode:
                 self.__load_wav_file(filename)
@@ -70,7 +77,12 @@ class Signal(QObject):
                 self.__load_complex_file(filename)
 
             self.filename = filename
-            self.noise_threshold = AutoInterpretation.detect_noise_level(self.iq_array.magnitudes)
+
+            default_noise_threshold = settings.read("default_noise_threshold", "automatic")
+            if default_noise_threshold == "automatic":
+                self.noise_threshold = AutoInterpretation.detect_noise_level(self.iq_array.magnitudes)
+            else:
+                self.noise_threshold = float(default_noise_threshold) / 100 * self.max_magnitude
         else:
             self.filename = ""
 
@@ -108,6 +120,7 @@ class Signal(QObject):
         self.iq_array = IQArray(None, np.float32, n=num_frames)
         if num_channels == 1:
             self.iq_array.real = np.multiply(1 / params["max"], np.subtract(data, params["center"]))
+            self.__already_demodulated = True
         elif num_channels == 2:
             self.iq_array.real = np.multiply(1 / params["max"], np.subtract(data[0::2], params["center"]))
             self.iq_array.imag = np.multiply(1 / params["max"], np.subtract(data[1::2], params["center"]))
@@ -126,6 +139,10 @@ class Signal(QObject):
         self.__load_complex_file(extracted_filename)
         os.remove(extracted_filename)
 
+    @property
+    def already_demodulated(self) -> bool:
+        return self.__already_demodulated
+
     @property
     def sample_rate(self):
         return self.__sample_rate
@@ -246,6 +263,18 @@ class Signal(QObject):
             if not self.block_protocol_update:
                 self.protocol_needs_update.emit()
 
+    @property
+    def costas_loop_bandwidth(self):
+        return self.__costas_loop_bandwidth
+
+    @costas_loop_bandwidth.setter
+    def costas_loop_bandwidth(self, value: float):
+        if self.__costas_loop_bandwidth != value:
+            self.__costas_loop_bandwidth = value
+            self._qad = None
+            if not self.block_protocol_update:
+                self.protocol_needs_update.emit()
+
     @property
     def message_length_divisor(self) -> int:
         return self.__message_length_divisor
@@ -281,24 +310,40 @@ class Signal(QObject):
             self._qad = None
             self.clear_parameter_cache()
             self._noise_threshold = value
-            self.noise_min_plot = -value
-            self.noise_max_plot = value
+
+            middle = 0.5*sum(IQArray.min_max_for_dtype(self.iq_array.dtype))
+            a = self.max_amplitude * value / self.max_magnitude
+            self.noise_min_plot = middle - a
+            self.noise_max_plot = middle + a
             self.noise_threshold_changed.emit()
             if not self.block_protocol_update:
                 self.protocol_needs_update.emit()
 
+    @property
+    def max_magnitude(self):
+        mi, ma = IQArray.min_max_for_dtype(self.iq_array.dtype)
+        return (2 * max(mi**2, ma**2))**0.5
+
+    @property
+    def max_amplitude(self):
+        mi, ma = IQArray.min_max_for_dtype(self.iq_array.dtype)
+        return 0.5 * (ma - mi)
+
     @property
     def noise_threshold_relative(self):
-        return self.noise_threshold / (self.iq_array.maximum**2.0 + self.iq_array.minimum**2.0)**0.5
+        return self.noise_threshold / self.max_magnitude
 
     @noise_threshold_relative.setter
     def noise_threshold_relative(self, value: float):
-        self.noise_threshold = value * (self.iq_array.maximum**2.0 + self.iq_array.minimum**2.0)**0.5
+        self.noise_threshold = value * self.max_magnitude
 
     @property
     def qad(self):
         if self._qad is None:
-            self._qad = self.quad_demod()
+            if self.already_demodulated:
+                self._qad = np.ascontiguousarray(self.real_plot_data, dtype=self.real_plot_data.dtype)
+            else:
+                self._qad = self.quad_demod()
 
         return self._qad
 
@@ -337,7 +382,12 @@ class Signal(QObject):
         QApplication.instance().restoreOverrideCursor()
 
     def quad_demod(self):
-        return signal_functions.afp_demod(self.iq_array.data, self.noise_threshold, self.modulation_type)
+        if self.noise_threshold < self.max_magnitude:
+            return signal_functions.afp_demod(self.iq_array.data, self.noise_threshold,
+                                              self.modulation_type, self.modulation_order,
+                                              self.costas_loop_bandwidth)
+        else:
+            return np.zeros(2, dtype=np.float32)
 
     def calc_relative_noise_threshold_from_range(self, noise_start: int, noise_end: int):
         num_digits = 4
@@ -367,7 +417,10 @@ class Signal(QObject):
         new_signal.__samples_per_symbol = self.samples_per_symbol
         new_signal.__bits_per_symbol = self.bits_per_symbol
         new_signal.__center = self.center
+        new_signal.wav_mode = self.wav_mode
+        new_signal.__already_demodulated = self.__already_demodulated
         new_signal.changed = True
+        new_signal.sample_rate = self.sample_rate
         return new_signal
 
     def get_thresholds_for_center(self, center: float, spacing=None):
@@ -476,7 +529,10 @@ class Signal(QObject):
     def filter_range(self, start: int, end: int, fir_filter: Filter):
         self.iq_array[start:end] = fir_filter.work(self.iq_array[start:end])
         self._qad[start:end] = signal_functions.afp_demod(self.iq_array[start:end],
-                                                          self.noise_threshold, self.modulation_type)
+                                                          self.noise_threshold,
+                                                          self.modulation_type,
+                                                          self.modulation_order,
+                                                          self.costas_loop_bandwidth)
         self.__invalidate_after_edit()
 
     def __invalidate_after_edit(self):
